"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var graphql_request_1 = require("graphql-request");
var lodash_1 = require("lodash");
var chalk_1 = require("chalk");
var introspectionQuery_1 = require("./introspectionQuery");
var debug = require('debug')('client');
var REMOTE_PROJECT_FRAGMENT = "\n  fragment RemoteProject on Project {\n    id\n    name\n    schema\n    alias\n    region\n    isEjected\n    projectDefinitionWithFileContent\n  }\n";
var MIGRATION_FRAGMENT = "\nfragment MigrationFragment on Migration {\n  revision\n  steps {\n    type\n    __typename\n    ... on CreateEnum {\n      name\n      ce_values: values\n    }\n    ... on CreateField {\n      model\n      name\n      cf_typeName: typeName\n      cf_isRequired: isRequired\n      cf_isList: isList\n      cf_isUnique: unique\n      cf_relation: relation\n      cf_defaultValue: default\n      cf_enum: enum\n    }\n    ... on CreateModel {\n      name\n    }\n    ... on CreateRelation {\n      name\n      leftModel\n      rightModel\n    }\n    ... on DeleteEnum {\n      name\n    }\n    ... on DeleteField {\n      model\n      name\n    }\n    ... on DeleteModel {\n      name\n    }\n    ... on DeleteRelation {\n      name\n    }\n    ... on UpdateEnum {\n      name\n      newName\n      values\n    }\n    ... on UpdateField {\n      model\n      name\n      newName\n      typeName\n      isRequired\n      isList\n      isUnique: unique\n      relation\n      default\n      enum\n    }\n    ... on UpdateModel {\n      name\n      um_newName: newName\n    }\n  }\n  hasBeenApplied\n}\n";
var Client = (function () {
    function Client(config, environment, out) {
        this.mocks = {};
        this.config = config;
        this.env = environment;
        this.out = out;
    }
    // always create a new client which points to the latest config for each request
    Client.prototype.getClient = function (cluster) {
        return new graphql_request_1.GraphQLClient(cluster.getDeployEndpoint(), {
            headers: {
                Authorization: "Bearer " + cluster.token,
            },
        });
    };
    Object.defineProperty(Client.prototype, "client", {
        get: function () {
            var _this = this;
            debug('choosing clusterEndpoint', this.env.activeCluster.getDeployEndpoint());
            var localClient = this.getClient(this.env.activeCluster);
            return {
                request: function (query, variables) { return __awaiter(_this, void 0, void 0, function () {
                    var e_1, message;
                    return __generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                debug('Sending query');
                                debug(query);
                                debug(variables);
                                _a.label = 1;
                            case 1:
                                _a.trys.push([1, 3, , 8]);
                                return [4 /*yield*/, localClient.request(query, variables)];
                            case 2: return [2 /*return*/, _a.sent()];
                            case 3:
                                e_1 = _a.sent();
                                if (!e_1.message.includes('No service with id')) return [3 /*break*/, 4];
                                message = e_1.response.errors[0].message;
                                this.out.error(message +
                                    " Please check if you are logged in to the right account.");
                                return [3 /*break*/, 7];
                            case 4:
                                if (!e_1.message.startsWith('No valid session')) return [3 /*break*/, 6];
                                return [4 /*yield*/, this.client.request(query, variables)];
                            case 5: 
                            // await this.auth.ensureAuth(true)
                            // try again with new token
                            return [2 /*return*/, _a.sent()];
                            case 6:
                                if (e_1.message.includes('ECONNREFUSED') &&
                                    (e_1.message.includes('localhost') || e_1.message.includes('127.0.0.1'))) {
                                    this.out.error("Could not connect to local cluster. Please use " + chalk_1.default.bold.green('graphcool local up') + " to start your local Graphcool cluster.");
                                }
                                else {
                                    throw e_1;
                                }
                                _a.label = 7;
                            case 7: return [3 /*break*/, 8];
                            case 8: return [2 /*return*/];
                        }
                    });
                }); },
            };
        },
        enumerable: true,
        configurable: true
    });
    // async getAccount(): Promise<AccountInfo> {
    //   const { viewer: { user } } = await this.client.request<{
    //     viewer: { user: AccountInfo }
    //   }>(`{
    //     viewer {
    //       user {
    //         email
    //         name
    //       }
    //     }
    //   }`)
    //   return user
    // }
    Client.prototype.introspect = function (serviceName, stageName, token) {
        return __awaiter(this, void 0, void 0, function () {
            var headers, client;
            return __generator(this, function (_a) {
                debug('introspecting', serviceName, stageName, token);
                headers = {};
                if (token) {
                    headers.Authorization = "Bearer " + token;
                }
                client = new graphql_request_1.GraphQLClient(this.env.activeCluster.getApiEndpoint(serviceName, stageName), {
                    headers: headers,
                });
                return [2 /*return*/, client.request(introspectionQuery_1.introspectionQuery)];
            });
        });
    };
    Client.prototype.download = function (serviceName, stage, exportData, token) {
        return __awaiter(this, void 0, void 0, function () {
            var endpoint, result, text;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        endpoint = this.env.activeCluster.getExportEndpoint(serviceName, stage);
                        debug("Downloading from " + endpoint);
                        debug(exportData);
                        return [4 /*yield*/, fetch(endpoint, {
                                method: 'post',
                                headers: {
                                    Authorization: "Bearer " + token,
                                    'Content-Type': 'application/json',
                                },
                                body: exportData,
                            })];
                    case 1:
                        result = _a.sent();
                        return [4 /*yield*/, result.text()];
                    case 2:
                        text = _a.sent();
                        try {
                            return [2 /*return*/, JSON.parse(text)];
                        }
                        catch (e) {
                            throw new Error(text);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    Client.prototype.upload = function (serviceName, stage, exportData, token) {
        return __awaiter(this, void 0, void 0, function () {
            var result, text;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, fetch(this.env.activeCluster.getImportEndpoint(serviceName, stage), {
                            method: 'post',
                            headers: {
                                Authorization: "Bearer " + token,
                                'Content-Type': 'application/json',
                            },
                            body: exportData,
                        })];
                    case 1:
                        result = _a.sent();
                        return [4 /*yield*/, result.text()];
                    case 2:
                        text = _a.sent();
                        try {
                            return [2 /*return*/, JSON.parse(text)];
                        }
                        catch (e) {
                            throw new Error(text);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    Client.prototype.reset = function (serviceName, stage, token) {
        return __awaiter(this, void 0, void 0, function () {
            var result, text;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, fetch(this.env.activeCluster.getApiEndpoint(serviceName, stage), {
                            method: 'post',
                            headers: {
                                Authorization: "Bearer " + token,
                                'Content-Type': 'application/json',
                            },
                            body: JSON.stringify({
                                query: "mutation {\n          resetData\n        }",
                            }),
                        })];
                    case 1:
                        result = _a.sent();
                        return [4 /*yield*/, result.text()];
                    case 2:
                        text = _a.sent();
                        try {
                            return [2 /*return*/, JSON.parse(text)];
                        }
                        catch (e) {
                            throw new Error(text);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    Client.prototype.addProject = function (name, stage, secrets) {
        return __awaiter(this, void 0, void 0, function () {
            var mutation, project;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        mutation = "      mutation addProject($name: String! $stage: String! $secrets: [String!]) {\n        addProject(input: {\n          name: $name,\n          stage: $stage\n          secrets: $secrets\n        }) {\n          project {\n            name\n          }\n        }\n      }\n      ";
                        return [4 /*yield*/, this.client.request(mutation, {
                                name: name,
                                stage: stage,
                                secrets: secrets,
                            })
                            // TODO set project definition, should be possibility in the addProject mutation
                        ];
                    case 1:
                        project = (_a.sent()).addProject.project;
                        // TODO set project definition, should be possibility in the addProject mutation
                        return [2 /*return*/, project];
                }
            });
        });
    };
    Client.prototype.getDeployUrl = function (projectId) {
        return __awaiter(this, void 0, void 0, function () {
            var mutation, url;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        mutation = "\n      mutation getUrl($projectId: String!) {\n        getTemporaryDeployUrl(\n          input: {\n          projectId: $projectId\n        }\n      ) {\n          url\n        }\n      }\n    ";
                        return [4 /*yield*/, this.client.request(mutation, { projectId: projectId })];
                    case 1:
                        url = (_a.sent()).getTemporaryDeployUrl.url;
                        return [2 /*return*/, url];
                }
            });
        });
    };
    Client.prototype.deploy = function (name, stage, types, dryRun, secrets) {
        return __awaiter(this, void 0, void 0, function () {
            var mutation, deploy;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        mutation = "      mutation($name: String!, $stage: String! $types: String! $dryRun: Boolean $secrets: [String!]) {\n        deploy(input: {\n          name: $name\n          stage: $stage\n          types: $types\n          dryRun: $dryRun\n          secrets: $secrets\n        }) {\n          errors {\n            type\n            field\n            description\n          }\n          migration {\n            ...MigrationFragment\n          }\n        }\n      }\n      " + MIGRATION_FRAGMENT + "\n    ";
                        return [4 /*yield*/, this.client.request(mutation, {
                                name: name,
                                stage: stage,
                                types: types,
                                dryRun: dryRun,
                                secrets: secrets,
                            })];
                    case 1:
                        deploy = (_a.sent()).deploy;
                        return [2 /*return*/, deploy];
                }
            });
        });
    };
    Client.prototype.listProjects = function () {
        return __awaiter(this, void 0, void 0, function () {
            var listProjects;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.client.request("\n      {\n        listProjects {\n          name\n          stage\n        }\n      }\n    ")];
                    case 1:
                        listProjects = (_a.sent()).listProjects;
                        return [2 /*return*/, listProjects];
                }
            });
        });
    };
    Client.prototype.getCluster = function (name, stage) {
        return __awaiter(this, void 0, void 0, function () {
            var foundClusters, _i, _a, cluster, client, project, e_2, clusterNames;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        foundClusters = [];
                        _i = 0, _a = this.env.clusters;
                        _b.label = 1;
                    case 1:
                        if (!(_i < _a.length)) return [3 /*break*/, 6];
                        cluster = _a[_i];
                        client = this.getClient(cluster);
                        _b.label = 2;
                    case 2:
                        _b.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, client.request("\n            query ($name: String! $stage: String!) {\n              project (name: $name stage: $stage) {\n                name\n                stage\n              }\n            }\n          ", {
                                name: name,
                                stage: stage,
                            })];
                    case 3:
                        project = (_b.sent()).project;
                        if (project) {
                            foundClusters.push(cluster);
                        }
                        return [3 /*break*/, 5];
                    case 4:
                        e_2 = _b.sent();
                        return [3 /*break*/, 5];
                    case 5:
                        _i++;
                        return [3 /*break*/, 1];
                    case 6:
                        if (foundClusters.length === 1) {
                            return [2 /*return*/, foundClusters[0]];
                        }
                        if (foundClusters.length > 1) {
                            clusterNames = foundClusters.map(function (c) { return c.name; }).join(', ');
                            throw new Error("The service name / stage combination \"" + name + "@" + stage + "\" is ambigious. It exists in clusters " + clusterNames);
                        }
                        return [2 /*return*/, null];
                }
            });
        });
    };
    Client.prototype.getClusterSafe = function (name, stage) {
        return __awaiter(this, void 0, void 0, function () {
            var cluster;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.getCluster(name, stage)];
                    case 1:
                        cluster = _a.sent();
                        if (!cluster) {
                            throw new Error("No cluster for \"" + name + "@" + stage + "\" found. Please make sure to deploy the stage " + stage);
                        }
                        return [2 /*return*/, cluster];
                }
            });
        });
    };
    Client.prototype.waitForLocalDocker = function (endpoint) {
        return __awaiter(this, void 0, void 0, function () {
            var valid, e_3;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        valid = false;
                        _a.label = 1;
                    case 1:
                        if (!!valid) return [3 /*break*/, 7];
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        debug('requesting', endpoint);
                        return [4 /*yield*/, graphql_request_1.request(endpoint, "\n            {\n              __schema {\n                directives {\n                  description\n                }\n              }\n            }\n            ")];
                    case 3:
                        _a.sent();
                        valid = true;
                        return [3 /*break*/, 5];
                    case 4:
                        e_3 = _a.sent();
                        debug(e_3);
                        valid = false;
                        return [3 /*break*/, 5];
                    case 5: return [4 /*yield*/, new Promise(function (r) { return setTimeout(r, 500); })];
                    case 6:
                        _a.sent();
                        return [3 /*break*/, 1];
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    Client.prototype.waitForMigration = function (name, stage, revision) {
        return __awaiter(this, void 0, void 0, function () {
            var migrationStatus;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!true) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.client.request("query ($name: String! $stage: String!) {\n              migrationStatus(name: $name stage: $stage) {\n                revision\n                hasBeenApplied\n              }\n            }\n        ", {
                                stage: stage,
                                name: name,
                            })];
                    case 1:
                        migrationStatus = (_a.sent()).migrationStatus;
                        if (migrationStatus.revision >= revision &&
                            migrationStatus.hasBeenApplied) {
                            return [2 /*return*/];
                        }
                        return [4 /*yield*/, new Promise(function (r) { return setTimeout(r, 500); })];
                    case 2:
                        _a.sent();
                        return [3 /*break*/, 0];
                    case 3: return [2 /*return*/];
                }
            });
        });
    };
    Client.prototype.authenticateCustomer = function (endpoint, token) {
        return __awaiter(this, void 0, void 0, function () {
            var result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, graphql_request_1.request(endpoint, "\n      mutation ($token: String!) {\n        authenticateCustomer(input: {\n          auth0IdToken: $token\n        }) {\n          token\n          user {\n            id\n          }\n        }\n      }\n      ", { token: token })];
                    case 1:
                        result = _a.sent();
                        return [2 /*return*/, result.authenticateCustomer];
                }
            });
        });
    };
    Client.prototype.getPats = function (projectId) {
        return __awaiter(this, void 0, void 0, function () {
            var permanentAuthTokens;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("\n      query ($projectId: ID!){\n        viewer {\n          project(id: $projectId) {\n            permanentAuthTokens {\n              edges {\n                node {\n                  id\n                  name\n                  token\n                }\n              }\n            }\n          }\n        }\n      }\n      ", { projectId: projectId })];
                    case 1:
                        permanentAuthTokens = (_a.sent()).viewer.project.permanentAuthTokens;
                        if (!permanentAuthTokens) {
                            return [2 /*return*/, []];
                        }
                        return [2 /*return*/, permanentAuthTokens.edges.map(function (edge) { return edge.node; })];
                }
            });
        });
    };
    Client.prototype.getFunctions = function (projectId) {
        return __awaiter(this, void 0, void 0, function () {
            var functions;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("\n      query ($projectId: ID!){\n        viewer {\n          project(id: $projectId) {\n            functions {\n              edges {\n                node {\n                  name\n                  id\n                  type\n                  stats {\n                    requestCount\n                    errorCount\n                  }\n                  __typename\n                }\n              }\n            }\n          }\n        }\n      }\n      ", { projectId: projectId })];
                    case 1:
                        functions = (_a.sent()).viewer.project.functions;
                        return [2 /*return*/, functions.edges.map(function (edge) { return edge.node; })];
                }
            });
        });
    };
    Client.prototype.getFunction = function (projectId, functionName) {
        return __awaiter(this, void 0, void 0, function () {
            var functions, normalizedFunctionName;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.getFunctions(projectId)];
                    case 1:
                        functions = _a.sent();
                        normalizedFunctionName = normalizeName(functionName);
                        return [2 /*return*/, (functions.find(function (fn) { return normalizeName(fn.name) === normalizedFunctionName; }) ||
                                null)];
                }
            });
        });
    };
    Client.prototype.getFunctionLogs = function (functionId, count) {
        if (count === void 0) { count = 1000; }
        return __awaiter(this, void 0, void 0, function () {
            var node;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("query ($id: ID!, $count: Int!) {\n      node(id: $id) {\n        ... on Function {\n          logs(last: $count) {\n            pageInfo {\n              endCursor\n            }\n            edges {\n              node {\n                id\n                requestId\n                duration\n                status\n                timestamp\n                message\n              }\n            }\n          }\n        }\n      }\n    }", { id: functionId, count: count })];
                    case 1:
                        node = (_a.sent()).node;
                        return [2 /*return*/, node && node.logs ? node.logs.edges.map(function (edge) { return edge.node; }) : null];
                }
            });
        });
    };
    Client.prototype.getAllFunctionLogs = function (projectId, count) {
        if (count === void 0) { count = 1000; }
        return __awaiter(this, void 0, void 0, function () {
            var project;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("\n        query ($id: ID!, $count: Int!) {\n          viewer {\n            project(id: $id) {\n              functions {\n                edges {\n                  node {\n                    logs(last: $count) {\n                      edges {\n                        node {\n                          id\n                          requestId\n                          duration\n                          status\n                          timestamp\n                          message\n                        }\n                      }\n                    }\n                  }\n                }\n              }\n            }\n          }\n        }\n        ", { id: projectId, count: count })];
                    case 1:
                        project = (_a.sent()).viewer.project;
                        return [2 /*return*/, project && project.functions && project.functions.edges
                                ? lodash_1.flatMap(project.functions.edges.map(function (functionEdge) {
                                    return functionEdge.node.logs.edges.map(function (logEdge) { return logEdge.node; });
                                }))
                                : null];
                }
            });
        });
    };
    Client.prototype.getProjectName = function (projectId) {
        return __awaiter(this, void 0, void 0, function () {
            var project;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("\n      query ($projectId: ID!){\n        viewer {\n          project(id: $projectId) {\n            name\n          }\n        }\n      }\n      ", { projectId: projectId })];
                    case 1:
                        project = (_a.sent()).viewer.project;
                        return [2 /*return*/, project.name];
                }
            });
        });
    };
    Client.prototype.deleteProjects = function (projectIds) {
        return __awaiter(this, void 0, void 0, function () {
            var inputArguments, singleMutations, header, body, mutation, variables, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        inputArguments = projectIds.reduce(function (prev, current, index) {
                            return prev + "$projectId" + index + ": String!" + (index < projectIds.length - 1 ? ', ' : '');
                        }, '');
                        singleMutations = projectIds.map(function (projectId, index) { return "\n      " + projectId + ": deleteProject(input:{\n          projectId: $projectId" + index + ",\n          clientMutationId: \"asd\"\n        }) {\n          deletedId\n      }"; });
                        header = "mutation (" + inputArguments + ") ";
                        body = singleMutations.join('\n');
                        mutation = header + " { \n" + body + " \n}";
                        variables = projectIds.reduce(function (prev, current, index) {
                            prev["projectId" + index] = current;
                            return prev;
                        }, {});
                        return [4 /*yield*/, this.client.request(mutation, variables)];
                    case 1:
                        result = _a.sent();
                        return [2 /*return*/, Object.keys(result).map(function (projectId) { return result[projectId].deletedId; })];
                }
            });
        });
    };
    Client.prototype.exportProjectData = function (projectId) {
        return __awaiter(this, void 0, void 0, function () {
            var exportData;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        return [4 /*yield*/, this.client.request("\n      mutation ($projectId: String!){\n        exportData(input:{\n          projectId: $projectId,\n          clientMutationId: \"asd\"\n        }) {\n          url\n        }\n      }\n    ", { projectId: projectId })];
                    case 1:
                        exportData = (_a.sent()).exportData;
                        return [2 /*return*/, exportData.url];
                }
            });
        });
    };
    return Client;
}());
exports.Client = Client;
// only make this available in test mode
// if (process.env.NODE_ENV === 'test') {
//   Client.prototype.mock = function({ request, response }) {
//     if (!this.mocks) {
//       this.mocks = {}
//     }
//     this.mocks[JSON.stringify(request, null, 2)] = response
//   }
// }
function normalizeName(name) {
    return name.toLowerCase().trim();
}
//# sourceMappingURL=Client.js.map